
/* 
 * File:   DSSOM.cpp
 * Author: hans
 * 
 * Created on 6 de Outubro de 2010, 17:41
 * OLD MWDSSOM
 */

#ifndef DSSOM_FUNCTIONS
#define DSSOM_FUNCTIONS

#include "DSSOM.h"
#include "DebugOut.h"
#include <math.h>


template <class TNeuron, class TSOMParameters>
DSSOM<TNeuron, TSOMParameters>::DSSOM() {
    this->section = "DSSOM";
    this->comments = "Dimension Selective Self-Organizing Map";
    this->callback = NULL;
}

template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::setParameters(uint numNeighbors, float alpha, float beta, float lambda2, float sigma, float lambda1, uint tmax) {
    this->parameters.numNeighbors = numNeighbors;
    this->parameters.alpha = alpha;
    this->parameters.lambda2 = lambda2;
    this->parameters.sigma = sigma;
    this->parameters.lambda1 = lambda1;
    this->parameters.tmax = tmax;
    this->parameters.beta = beta;
}

template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::initialize(uint somRows, uint somCols, uint inputSize) {
    this->somRows = somRows;
    this->somCols = somCols;
    this->inputSize = inputSize;
    resetTime();

    this->neurons.size(somRows, somCols);
    this->activationMap.size(somRows, somCols);
    relevance.size(inputSize);
    relevance.fill(1);

    reRandomizeWeights();
}


template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::reRandomizeWeights() {

    for (uint r = 0; r < this->neurons.rows(); r++)
        for (uint c = 0; c < this->neurons.cols(); c++) {
            this->neurons[r][c].set(r, c);
            this->neurons[r][c].setupWeights(this->inputSize);
        }

    this->activationMap.fill(0);
}

template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::updateNeuron(TNeuron &neuron, MatVector<float> &data, float alpha, float phi) {


    for (uint i = 0; i < data.size(); i++) {
        //update neuron weights
        float distance = fabs(data[i] - neuron.weights[i]);
        neuron.avgDistance[i] = phi * beta_t * distance + (1 - phi * beta_t) * neuron.avgDistance[i];
    }

    float max = neuron.avgDistance.max();

    //update neuron fs weights
    for (uint i = 0; i < data.size(); i++) {
        if (max != 0)
            neuron.dsWeights[i] = 1 - (neuron.avgDistance[i] / max);
        else
            neuron.dsWeights[i] = 1;

        if (neuron.dsWeights[i] < this->parameters.minDW)
            neuron.dsWeights[i] = this->parameters.minDW;
    }

    //update neuron weights
    for (uint i = 0; i < data.size(); i++) {

        neuron.weights[i] = neuron.weights[i] + phi * alpha * (data[i] - neuron.weights[i]);
    }
}

template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::resetTime() {
    this->t = 0;
    //alpha_t = ((parameters.tmax-t)*parameters.alpha)/parameters.tmax;  //linear decay
    this->alpha_t = this->parameters.alpha * exp(-(float) this->t / (this->parameters.tmax * this->parameters.lambda2)); //exponencial decay
    this->beta_t = this->parameters.beta;
    this->sigma_t = this->parameters.sigma * exp(-((int) this->t + 1) / (this->parameters.tmax * this->parameters.lambda1));
}

template <class TNeuron, class TSOMParameters>
bool DSSOM<TNeuron, TSOMParameters>::incTime() {
    this->t++;
    if (this->t < this->parameters.tmax) {
        //alpha_t = ((parameters.tmax-t)*parameters.alpha)/parameters.tmax;  //linear decay
        this->alpha_t = this->parameters.alpha * exp(-(float) this->t / (this->parameters.tmax * this->parameters.lambda2)); //exponencial decay
        if (this->alpha_t < 0.00001) this->alpha_t = 0.00001;
        this->sigma_t = this->parameters.sigma * exp(-((int) this->t + 1) / (this->parameters.tmax * this->parameters.lambda1));
    }

    return this->t < this->parameters.tmax;
}


template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::resetRelevances() {
    relevance.fill(1); //Set all features as full relevant
}

template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::updateRelevances(TNeuron &bmu) {
   //Reduce the relevance of features already considered by current bmu
   for (uint i = 0; i < relevance.size(); i++)
      relevance[i] = relevance[i]*(1-bmu.dsWeights[i]);
      //relevance[r][c] = relevance[r][c]*exp(-bmu.dsWeights[r][c]*7);
}

template <class TNeuron, class TSOMParameters>
const MatVector<float> DSSOM<TNeuron, TSOMParameters>::getRelevance() {
    return relevance;
}

template <class TNeuron, class TSOMParameters>
float DSSOM<TNeuron, TSOMParameters>::findFirstBMU(const MatVector<float> &data, TNeuron &bmu) {
       
    uint r,c;
    for (r=0; r<this->neurons.rows(); r++)
    for (c=0; c<this->neurons.cols(); c++)
        this->neurons[r][c].canWin = true;

    this->presentPaternToSOM(data);

    this->activationMap.max(r,c);
    bmu = this->neurons[r][c];

    return bmu.activation;
}

template <class TNeuron, class TSOMParameters>
float DSSOM<TNeuron, TSOMParameters>::findNextBMU(const MatVector<float> &data, TNeuron &bmu) {

    this->neurons[bmu.r][bmu.c].canWin = false; //A winner neuron can not win again for this input

    uint r,c;
    /*     
    for (r=0; r<neurons.rows(); r++)
    for (c=0; c<neurons.cols(); c++)
    {
        //activationMap[r][c] = neurons[r][c].activation = presentPaternToNeuronAgain(data, neurons[r][c]);
        activationMap[r][c] = neurons[r][c].activation = presentPaternToNeuron(data, neurons[r][c]);
    }/**/

    while (true) {
     this->activationMap.max(r,c);
     bmu = this->neurons[r][c];

     if (this->activationMap[r][c] == 0)
         break;

     if (bmu.canWin)
         break;
     else
         this->activationMap[r][c] = 0;
    };

    return bmu.activation;
}

template <class TNeuron, class TSOMParameters>
float DSSOM<TNeuron, TSOMParameters>::findNextBMU(TNeuron &bmu)
{
    uint r,c;
    float difference;
    TNeuron prevBmu = bmu;
    MatMatrix<float> dsWeights;

//    dbgOut(1) << "Winner0 [" << bmu.r << "," << bmu.c << "]: " << featuresToString(bmu.dsWeights) << endl;
    do {
        //desativa vencedor atual
        this->activationMap[bmu.r][bmu.c] = this->neurons[bmu.r][bmu.c].activation = 0;
        //pega o próximo vencedor
        this->activationMap.max(r,c);
        bmu = this->neurons[r][c];

        //Computa a distancia do pesso do vencedor atual para o vencedor original
        dsWeights = bmu.dsWeights;
        dsWeights.mult(-1);
        dsWeights.add(prevBmu.dsWeights);
        dsWeights.mqrt();
        difference = dsWeights.sum()/(dsWeights.rows()*dsWeights.cols());

        //Se a distancia > 0.2 encontrou vencedor de outro grupo
        if (difference>0.2) break;

        //Finaliza a busca se encontrar algum vencedor com ativação igual a zero
      } while (bmu.activation>0);
    //} while (bmu.activation>0.5*prevBmu.activation);

 //   if (bmu.activation > 0.5*prevBmu.activation)
       // dbgOut(1) << "WinnerX [" << bmu.r << "," << bmu.c << "]: "  << featuresToString(bmu.dsWeights) << endl;

      return bmu.activation;
}

template <class TNeuron, class TSOMParameters>
float DSSOM<TNeuron, TSOMParameters>::presentPaternToNeuronAgain(const MatVector<float> &data, TNeuron &neuron)
{
    float distance = 0;

    for (uint i = 0; i < data.size(); i++) {
            distance += (neuron.dsWeights[i]*relevance[i])*qrt(fabs(data[i] - neuron.weights[i]));
        }
    
    /**/
    float sumq=0;
    for (uint i = 0; i < data.size(); i++) {

            sumq += qrt(neuron.dsWeights[i]-relevance[i]);
        }
    sumq = qrt(sumq);
    return (sumq/(sumq+distance+0.0000001));
    /**/
    
    //return (neuron.dsWeights.sum()/(neuron.dsWeights.sum()+distance*distance+0.0000001));
}

template <class TNeuron, class TSOMParameters>
float DSSOM<TNeuron, TSOMParameters>::presentPaternToNeuron(const MatVector<float> &data, TNeuron &neuron)
{
    float distance = 0;

    for (uint i = 0; i < data.size(); i++) {
            distance += (neuron.dsWeights[i])*qrt(fabs(data[i] - neuron.weights[i]));
        }
    
    /**/
    float sumq=0;
    for (uint i = 0; i < data.size(); i++) {
            sumq  += fabs(neuron.dsWeights[i]);
        }
    sumq = sqrt(sumq);
    //return (sumq/(sumq+distance+0.0000001));
    return (exp(-(distance)));
    /**/
    
    //return (neuron.dsWeights.sum()/(neuron.dsWeights.sum()+distance*distance+0.0000001));
}

template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::deactivateNeighbors(TNeuron &neuron)
{
    uint rmin = max((int)neuron.r-(int)this->parameters.numNeighbors, 0);
    uint rmax = min((int)neuron.r+(int)this->parameters.numNeighbors, (int)this->neurons.rows()-1);
    uint cmin = max((int)neuron.c-(int)this->parameters.numNeighbors, 0);
    uint cmax = min((int)neuron.c+(int)this->parameters.numNeighbors, (int)this->neurons.cols()-1);

    for (uint nr = rmin; nr<=rmax; nr++) //for each row neighbor of neuron
    for (uint nc = cmin; nc<=cmax; nc++) //for each col neighbor of neuron
    {
        this->activationMap[nr][nc] = this->neurons[nr][nc].activation = 0;
    }
}

template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::train(MatMatrix<float> &trainingData)
{
    TNeuron bmu;

    for (resetTime(); this->t<this->parameters.tmax; incTime())
    {
        trainStep(trainingData);
        if (this->callback!=NULL)
            if (this->callback(*this)) break;
    }
    relevance.fill(1);
}

template <class TNeuron, class TSOMParameters>
void DSSOM<TNeuron, TSOMParameters>::trainStep(MatMatrix<float> &trainingData)
{
    TNeuron bmu;

    for (uint i=0; i<trainingData.rows(); i++)
    {
        resetRelevances();
        
        int v = (rand() / (double) RAND_MAX)*(trainingData.rows() - 1);
        MatVector<float> sample;
        trainingData.getRow(v, sample);
        findFirstBMU(sample, bmu);

        this->updateNeigbors(bmu, sample, this->alpha_t);

        float alpha_t2 = this->alpha_t;
        for (uint w=1; w<this->parameters.numWinners; w++)
        {
           updateRelevances(bmu);
           if (relevance.max()<=this->parameters.epsilonRho) break;
           alpha_t2 = alpha_t2*this->sigma_t/this->parameters.sigma;

           findNextBMU(sample, bmu);

           /*float max=0;
           for (int r=0; r< relevance.cols(); r++) {
               float rds = relevance[0][r]*bmu.dsWeights[0][r];
               if (rds>max)
                   max = rds;
           }
           if (max>parameters.epsilonRho)*/
           float beta_t2 = beta_t;
           beta_t = 0;
           this->updateNeigbors(bmu, sample, alpha_t2);
           beta_t = beta_t2;
        }

        //dbgOut(1) << i << ": " << relevance.max() << endl;
        
        if (this->callback!=NULL)
            if (this->callback(*this)) break;

        //dbgOut(0) << (*trainingData[v]).toString() << endl;
    }
    resetRelevances();
}
#endif