/* 
 * File:   main.cpp
 * Author: hans
 *
 * Created on 11 de Outubro de 2010, 07:25
 */

#include <stdlib.h>
#include <fstream>
#include <algorithm>
#include <time.h>
#include <limits>
#include <unistd.h>
#include "MatMatrix.h"
#include "MatVector.h"
#include "Defines.h"
#include "DebugOut.h"
#include "Parameters.h"
#include "ArffData.h"
#include "ClusteringSOM.h"
#include "DSSOM.h"
#include "DSNeuron.h"

using namespace std;

int main(int argc, char** argv) {

    int seed = 0;;
    dbgThreshold(1);

    DSSOM<DSNeuron, DSSOMParameters> som;

    som.parameters.NX = 1;
    som.parameters.NY = 10;
    som.parameters.numNeighbors = 1; //Num numNeighbors: City block distance to the farthest neigbors
    som.parameters.sigma = 0.9; //Sigma0: Neighborhood function standard deviation
    som.parameters.lambda1 = 0.05; //Neighborhood function decay with time
    som.parameters.alpha = 0.001; //Alpha0: Learning rate
    som.parameters.lambda2 = 0.9; //Learning rate decay with time
    som.parameters.tmax = 2000; //Maximum number of iterations
    
    som.parameters.beta = 0.001; //Beta0: Distance average rate
    som.parameters.minDW = 0.001; //Minimum value for dsWeights
    som.parameters.numWinners = 1; //Maximun number of Winners per input stimulus
    som.parameters.epsilonRho = 0.7; //Minimum relevance required to search for another winner
    som.parameters.outliersThreshold = 0; //Minimum activation to consider outlier
    
    string filename = "";

    int c;
    while ((c = getopt(argc, argv, "f:n:t:x:y:o:g:s:l:a:d:b:m:w:e:r:")) != -1)
        switch (c) {
            case 'f':
                filename.assign(optarg);
                break;
            case 'n':
            case 't':                
                som.parameters.tmax = atoi(optarg);
                break;                
            case 'x':
                som.parameters.NX = atoi(optarg);
                break;
            case 'y':
                som.parameters.NY = atoi(optarg);
                break;
            case 'o':
                som.parameters.outliersThreshold = atof(optarg);
                break;
            case 'g':
                som.parameters.numNeighbors = atoi(optarg);
                break;
            case 's':
                som.parameters.sigma = atof(optarg);
                break;
            case 'l':
                som.parameters.lambda1 = atof(optarg);
                break;
            case 'a':
                som.parameters.alpha = atof(optarg);
                break;
            case 'd':
                som.parameters.lambda2 = atof(optarg);
                break;
            case 'b':
                som.parameters.beta = atof(optarg);
                break;
            case 'm':
                som.parameters.minDW = atof(optarg);
                break;
            case 'w':
                som.parameters.numWinners = atoi(optarg);
                break;
            case 'e':
                som.parameters.epsilonRho = atof(optarg);
                break;
            case 'r':
                seed = atoi(optarg);
                break;
            case '?':
                if (optopt == 'f')
                    fprintf(stderr, "Option -%c requires an argument.\n", optopt);
                else if (isprint(optopt))
                    fprintf(stderr, "Unknown option `-%c'.\n", optopt);
                else
                    fprintf(stderr,
                        "Unknown option character `\\x%x'.\n",
                        optopt);
                return 1;
        }

    if (filename == "") {
        dbgOut(0) << "option -f filename is required" << endl;
        return -1;
    }

    dbgOut(1) << "Running DSSOM for file: " << filename << endl;
    srand(seed);
    cout << som.parameters << endl;

    ClusteringDSSOM clusteringSOM(&som);
    
    if (clusteringSOM.readFile(filename)){
        clusteringSOM.trainSOM(som.parameters.tmax);
        clusteringSOM.writeClusterResults(filename + ".results");
        dbgOut(1) << "Done." << endl;
    }
}